{%- block doctype -%}
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
{%- endblock %}
{%- set reldelim1 = reldelim1 is not defined and ' &raquo;' or reldelim1 %}
{%- set reldelim2 = reldelim2 is not defined and ' |' or reldelim2 %}

<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
{{ metatags }}
<title>{{ title|striptags }}</title>
<link rel="stylesheet" href="{{ pathto('_static/', 1) }}index.css" type="text/css" />
{%- block extrahead %} {% endblock %}
</head>
<body class="homepage">
<div class="document">
<div class="body">
{{ body }}
</div>
</div>
</body>
</html>
