<table class="contentstable" width=50%><tr>
<td width="50%">
<p class="biglink"><a class="biglink" href="{{ pathto("contents") }}">{{ _('Contents') }}</a><br>
<span class="linkdescr">{{ _('lists all sections and subsections') }}</span></p>
<p class="biglink"><a class="biglink" href="{{ pathto("search") }}">{{ _('Search Page') }}</a><br>
<span class="linkdescr">{{ _('search this documentation') }}</span></p>
</td><td width="50%">
<p class="biglink"><a class="biglink" href="{{ pathto("genindex") }}">{{ _('Index') }}</a><br>
<span class="linkdescr">{{ _('all functions, classes, terms') }}</span></p>
<p class="biglink"><a class="biglink" href="{{ pathto("modindex") }}">{{ _('Module Index') }}</a><br>
<span class="linkdescr">{{ _('quick access to all modules') }}</span></p>
</td></tr>
</table>
